from abc import ABCMeta, abstractmethod
from decimal import *
import re


class NumberRepresentation(metaclass=ABCMeta):

    def __init__(self, base: int, max_precision=10):
        """
        Creates number representation with source or target base
        :param base: source or target base
        :type base: int
        :param max_precision: number of decimal places to go
        :type max_precision: int
        """

        self.base = base
        self.precision = max_precision
        self.positive = True
        self.integ = Decimal('0')
        self.fract = Decimal('0.0')

    @abstractmethod
    def from_base(self, value: str, delimiter="", rounding=True):
        """
        Converts from the specified base to internal number structure
        :param value: string representation of the number
        :type value: str
        :param delimiter: text delimiter to use between characters if necessary (defaults to empty)
        :type delimiter: str
        :param rounding: round to precision
        :type rounding:bool
        :return: arbitrary precision number
        :rtype: Decimal
        """
        pass

    @abstractmethod
    def to_base(self, value: Decimal, delimiter="", rounding=True):
        """
        Converts from the internal number structure to a list of strings
        :param value: arbitrary precision number
        :type value: Decimal
        :param delimiter: text delimiter to use between characters if necessary (defaults to empty)
        :type delimiter: str
        :param rounding: round to precision
        :type rounding:bool
        :return: number represented in new base as a list
        :rtype: list of str
        """
        pass

    def to_base_string(self, value: Decimal):
        """
        Converts from the internal number structure to a single string
        :param value: arbitrary precision number
        :type value: int
        :return: number represented in new base as a string
        :rtype: str
        """
        return self.delimiter.join(self.to_base(value))

    def clean_up_number_string(self, value: str, delimiter=""):
        """
        spring cleaning on the string
        :param value: string representation of the number
        :type value: str
        :param delimiter: delimiter between digits, defaults to ""
        :type delimiter: str
        :return: cleaned string, split into integer and fractional parts
        :rtype: tuple of str
        """
        value = value.strip().upper()

        if delimiter != "":
            parts = value.split(delimiter)
            value = ""
            for n in range(len(parts)):
                parts[n] = re.sub(r'\s*', '', parts[n])

            value = delimiter.join(parts)
            value = re.sub(r'\s*\.\s*', '.', value)
        else:
            value = re.sub(r'\s*', '', value)

        if value[0:1] == '-.':
            value = "0" + value[1:]
        if value[0] == '.':
            value = "0" + value
        parts = value.split('.', 2)
        i = parts[0]
        if len(parts) > 1:
            f = parts[1]
        else:
            f = ""
        self.positive = i[0] != '-'
        if not self.positive:
            i = i[1:]
        return i, f

    def clean_up_number(self, value: Decimal):
        """
        spring cleaning on the internal value
        :param value: arbitrary precision representation of the number
        :return: cleaned value, split into integer and fractional parts
        :rtype: tuple of Decimal
        """
        (istr, fstr) = self.clean_up_number_string(str(value))

        i = Decimal(istr)
        f = Decimal("0." + fstr)

        return i, f

    @staticmethod
    def frac(value: Decimal):
        return Decimal(abs(value) - int(abs(value))).copy_sign(value)


from decimal import *
from .number_representation import NumberRepresentation


class LargeBaseNumber(NumberRepresentation):

    def __init__(self, base: int, max_precision=10):
        super(LargeBaseNumber, self).__init__(base, max_precision)

    def from_base(self, value: str, delimiter="", rounding=True):
        (self.integ, self.fract) = self.clean_up_number_string(value, delimiter)
        eprec = ((len(self.fract) + len(self.integ)) * 2)
        extcontext = Context(prec=eprec, Emin=eprec*-2, Emax=eprec*2)
        setcontext(extcontext)

        if self.base > 36 and delimiter == "":
            raise Exception("bases larger than 36 MUST specify the delimiter between digits")

        # integer part
        n = Decimal('0')
        if delimiter == "":
            iparts = str(self.integ)
        else:
            iparts = str(self.integ).split(delimiter)
        for i in range(len(iparts)):
            n *= self.base
            n += Decimal(iparts[i])

        # fraction part
        f = Decimal('0')
        if delimiter == "":
            fparts = str(self.fract)
        else:
            fparts = str(self.fract).split(delimiter)
        if self.fract is not "" and self.fract is not "0" and self.fract is not "0.0":
            for i in range(len(fparts)):
                f *= Decimal(self.base)
                f += Decimal(fparts[i])
            invf = Decimal(pow(Decimal(self.base), Decimal(0 - (len(fparts)))))
            f *= invf

        combined_value = Decimal('0')
        if rounding:
            with localcontext() as ctx:
                ctx.prec = self.precision + len(str(Decimal(n)))
                combined_value += n
                combined_value += f
        else:
            with localcontext() as ctx:
                ctx.prec = self.precision + len(str(Decimal(n)))
                ctx.rounding = ROUND_DOWN
                combined_value += n
                combined_value += f

        if not self.positive:
            combined_value = 0 - combined_value

        return combined_value

    def to_base(self, value: Decimal, delimiter="", rounding=True):
        if value == Decimal('0'):
            return "0"

        (self.integ, self.fract) = self.clean_up_number(value)
        eprec = ((len(str(self.fract)) + len(str(self.integ))) * 2)
        extcontext = Context(prec=eprec, Emin=eprec*-2, Emax=eprec*2)
        setcontext(extcontext)

        # integer part
        remainder = self.integ
        idigits = []
        while remainder > 0:
            tmp = remainder % self.base
            idigits.insert(0, str(tmp))
            remainder = (remainder - tmp) / self.base

        # fraction part
        remainder = self.fract
        fdigits = []
        while remainder > 0 and len(fdigits) <= self.precision:
            remainder *= self.base
            tmp = int(remainder)
            remainder = self.frac(remainder)
            fdigits.append(str(tmp))

        if remainder != 0 and rounding:
            if int(fdigits[-1]) >= self.base/2.0:
                fdigits = self.round(fdigits)
            else:
                # if it would be rounding down, we still need to strip off the
                # last character that was used to determine rounding
                fdigits = fdigits[:-1]
        elif remainder != 0 and not rounding:
            # if not rounding, strip off the last digit that would have been
            # used for rounding
            fdigits = fdigits[:-1]

        numstr = delimiter.join(idigits)
        if len(numstr) == 0:
            numstr = "0"

        if len(fdigits) > 0:
            numstr += "."
            numstr += delimiter.join(fdigits)

        if not self.positive:
            numstr = "-" + numstr

        return numstr

    def round(self, value: [str]):
        """
        rounds any number, not just non-decimal
        :param value: string of number
        :type value: list of str
        :return: string, new rounded number
        """
        if int(value[-1]) >= self.base/2.0:
            v = int(value[-2]) + 1
            if v < self.base:
                value[-2] = str(v)
                value = value[:-1]
            else:
                value = self.round(value[:-1])

        return value


# Copyright (c) 2015 David Woods. All rights reserved.
#
# License:
#   The MIT License
#
# See the file "LICENSE" for terms & conditions for usage, and a DISCLAIMER OF
# ALL WARRANTIES.
#

"""
    ALL YOUR BASE:

    All Your Base is a python library for converting number strings from any base to number strings of any other base.

    This library was created to make the following improvements on existing base conversion libraries out there:

    - Can convert both integers and floats
    - Uses Decimal package to allow for arbitrary precision / number of digits
    - Uses Decimal package to avoid binary rounding errors
    - Is not limited to base 36, 62, 64 due to available characters. Can convert to/from any integer base from 2 to
      whatever you like. (higher bases use delimited decimal format)



    INSTALLATION:

    Pip is preferred::

    > pip install allyourbase

    You can also install manually::

    > python setup.py install



    USAGE:

    > from allyourbase import BaseConvert
    > from decimal import Decimal
    >
    > numrep = BaseConvert(source_base=34)
    > numrep.decode("GKFT9")
    Decimal('22185791')
    >
    > numrep = BaseConvert(10, 16, max_precision=8)
    > numrep.encode(Decimal('4991.58791200'))
    "137F.9681669D"
    >
    > numrep = BaseConvert(source_base=16, destination_base=42)
    > numrep.convert("137F.9681669D", destination_delimiter=":", destination_max_precision=4)
    "2:34:35.24:29:3:9"

"""

from .standard_number import StandardNumber
from .large_base_number import LargeBaseNumber
from decimal import Decimal


class BaseConvert(object):
    """
    convert numbers (in string representation) from one base to another
    """

    def __init__(self, source_base=10, destination_base=16,
                 source_delimiter="", destination_delimiter="",
                 max_precision=10, rounding=True):
        """
        :param source_base: FROM base
        :type source_base: int
        :param destination_base: TO base
        :type destination_base: int
        :param source_delimiter: delimiter string for input
        :type source_delimiter: str
        :param destination_delimiter: delimiter string for output
        :type destination_delimiter: str
        :param max_precision: precision / decimal places (default 10)
        :type max_precision: int
        :param rounding: round (True) or just truncate (False)?
        :type rounding: bool
        """

        self.from_base = source_base
        self.to_base = destination_base
        self.source_delim = source_delimiter
        self.dest_delim = destination_delimiter
        self.precision = max_precision
        self.rounding = rounding

    def encode(self, value, base=None, delimiter=None, max_precision=None, rounding=None):
        """
        :param base: TO base
        :type base: int
        :param value: input value. Decimal or str is preferred for precision. Int can also be used.
        :param delimiter: delimiter string for output, used for bases > 36
        :type delimiter: str
        :param max_precision: precision / decimal places (default 10)
        :type max_precision: int
        :param rounding: round (True) or just truncate (False)?
        :type rounding: bool
        :return: string representing value in target base
        :rtype: str
        """
        if base is None:
            base = self.to_base
        if delimiter is None:
            delimiter = self.dest_delim
        if max_precision is None:
            max_precision = self.precision
        if rounding is None:
            rounding = self.rounding

        torep = self.__get_number_representation(base, max_precision)
        return torep.to_base(Decimal(value), delimiter, rounding)

    def decode(self, value: str, base=None, delimiter=None, max_precision=None, rounding=None):
        """
        :param base: FROM base
        :type base: int
        :param value: input value
        :type value: str
        :param delimiter: delimiter string for input, needed for bases > 36
        :type delimiter: str
        :param max_precision: precision / decimal places (default 10)
        :type max_precision: int
        :param rounding: round (True) or just truncate (False)?
        :type rounding: bool
        :return: coverted value to base10
        :rtype: Decimal
        """
        if base is None:
            base = self.from_base
        if delimiter is None:
            delimiter = self.source_delim
        if max_precision is None:
            max_precision = self.precision
        if rounding is None:
            rounding = self.rounding

        fromrep = self.__get_number_representation(base, max_precision)
        return fromrep.from_base(value, delimiter, rounding)

    def convert(self, value: str, source_base=None, destination_base=None, source_delimiter=None,
                destination_delimiter=None, source_max_precision=None, destination_max_precision=None, rounding=None):
        """
        :param source_base: FROM base
        :type source_base: int
        :param destination_base: TO base
        :type destination_base: int
        :param value: input value
        :type value: str
        :param source_delimiter: delimiter string for input, needed for bases > 36
        :type source_delimiter: str
        :param destination_delimiter: delimiter string for output, used for bases > 36
        :type destination_delimiter: str
        :param source_max_precision: precision / decimal places (default 10) for decoding
        :type source_max_precision: int
        :param destination_max_precision: precision / decimal places (default 10) for encoding
        :type destination_max_precision: int
        :param rounding: round (True) or just truncate (False)?
        :type rounding: bool
        :return: string of output
        :rtype: str
        """

        return self.encode(self.decode(str(value), base=source_base, delimiter=source_delimiter,
                                       max_precision=source_max_precision, rounding=rounding),
                           base=destination_base, delimiter=destination_delimiter,
                           max_precision=destination_max_precision, rounding=rounding)

    def __get_number_representation(self, base=None, max_precision=None):
        """
        Factory for creating number representations
        :param base: FROM or TO base
        :type base: int
        :param max_precision: precision / decimal places (default 10)
        :type max_precision: int
        :return: NumberRepresentation inherited class
        :rtype: _NumberRepresentation
        """
        if base is None:
            base = self.from_base
        if max_precision is None:
            max_precision = self.precision

        if base != int(base):
            raise Exception("illegal base. must be an integer greater than 1.")
        if base > 36:
            return LargeBaseNumber(base, max_precision)
        elif 2 <= base <= 36:
            return StandardNumber(base, max_precision)
        else:
            raise Exception("illegal base. must be an integer greater than 1.")